/*	Interface to TextEdit, so application is separated from TextEdit and
	can use a different text editing library, such as TE32K, which supports
	files larger than 32K.
	
	94/01/16 aih for consistency with the rest of the libraries made
					 constants uppercase
	94/01/12 aih improved check for decimal point in floating point hook
					 added maxlength field
	94/01/02 aih added wrap to view flag
	93/12/19 aih added TE32K
	93/12/18 aih added text hooks
	93/12/17 aih added LRectType
	93/11/29 aih added frame field
	93/11/05 aih created */

#include <Packages.h>
#include <Script.h>
#include "DrawLib.h"
#include "KeyLib.h"
#include "LowMemLib.h"
#include "MathLib.h"
#include "MemoryLib.h"
#include "RectangleLib.h"
#include "ScrapLib.h"
#include "TextEditLib.h"
#include "TextEdit32KLib.h"
#include "TextLib.h"
#include "TextMacros.h"
#include "WindowLib.h"

/* When the mouse is dragged outside of the text it will cause the text
	to scroll. On fast machines this can be problematic for a user, so if
	the mouse is within TX_SLOW_MARGIN pixels of the outside of the view
	rectangle a delay of TX_SLOW_DELAY ticks is inserted before the text
	is scrolled. Moving the mouse further away from the view rectangle
	disables this delay, causing the text to scroll at its maximum speed.
	From the TextEditControl by Eric Soldan, distributed by MacDTS on the
	7.0 Beta CD-ROM of February 1990. */
#define TX_SLOW_MARGIN	(12)
#define TX_SLOW_DELAY	(9)

static TextHandle gClickText;	/* for click loop */
static RgnHandle gClickRgn;	/* for click loop */

/* true if a valid text kind */
static Boolean TxKindValid(TextKindType kind)
{
	return(0 <= kind && kind < TX_LAST_KIND);
}

/* accessing text edit record */
static Boolean TxIsEdit(TextHandle text)
{
	require(TxValid(text));
	return((**text).kind != TX_EXTENDED_KIND);
}

/* true if can paste the contents of the scrap */
static Boolean CanPaste(TextHandle text)
{
	require(TxValid(text));
	return(! (**text).hooks.paste || (**text).hooks.paste(text, (**text).hooks.data));
}

/* validation */

static Boolean RangeValid(TextIntType n, TextIntType nmin, TextIntType nmax)
{
	return(nmin <= n && n <= nmax);
}

Boolean TxValid(TextHandle text)
{
	Rect view;
	LRectType dest;
	static Boolean validating;
	
	if (validating) return(true);
	validating = true;
	if (! HandleValidSize(text, sizeof(TextType))) goto invalid;
	if (! HandleValid((**text).te)) goto invalid;
	if (! HandleValid(TxText(text))) goto invalid;
	if (! TxKindValid((**text).kind)) goto invalid;
	if (HandleSize(TxText(text)) < TxLength(text)) goto invalid;
	if (! RangeValid(TxLength(text), 0, TxMaxLength(text))) goto invalid;
	if (! RangeValid(TxLines(text), 0, TxMaxLines(text))) goto invalid;
	if (! RangeValid(TxSelStart(text), 0, TxLength(text))) goto invalid;
	if (! RangeValid(TxSelEnd(text), 0, TxLength(text))) goto invalid;
	TxView(text, &view); if (! RectValid(&view)) goto invalid;
	TxDestination(text, &dest); if (! LRectValid(&dest)) goto invalid;
	/*
		These conditions must hold, but may be temporarily
		invalidated while changing the destination and view
		rectangles.
		if (dest.top > view.top) return(false);
		if (dest.left > view.left) return(false);
	*/
	validating = false;
	return(true);
invalid:
	validating = false;
	return(false);
}

Boolean TxLineValid(TextHandle text, TextIntType line)
{
	return(0 <= line && line <= TxLines(text));
}

Boolean TxOffsetValid(TextHandle text, TextIntType offset)
{
	return(0 <= offset && offset <= TxLength(text));
}

/* allocation */

void TxEnd(TextHandle text)
{
	if (text) {
		if ((**text).te) {
			WinTabUnregister(TxPort(text), text);
			WinUnregister(TxPort(text), text);
			TEMDispose(text);
		}
		FrameEnd((**text).frame);
		HandleEnd(text);
		text = NULL;
	}
	ensure(! TxValid(text));
}

static TE32KHandle TE32KBegin(const Rect *view)
{
	static Boolean initialized = false;
	LongRect lview;
	TE32KHandle te;
	
	if (! initialized) {
		TE32KInit();
		initialized = true;
	}
	RectToLongRect((Rect *) view, &lview);
	te = TE32KNew(&lview, &lview);
	FailNIL(te);
	return(te);
}

static TEHandle TEBegin(const Rect *view, Boolean styled)
{
	TEHandle te;
	
	if (styled)
		te = TEStylNew(view, view);
	else
		te = TENew(view, view);
	FailNIL(te);
	TECalText(te); /* fix a bug described in TN#131 */
	return(te);
}

TextHandle TxBegin(WindowPtr window, const Rect *bounds, TextKindType kind)
{
	volatile TextHandle text = NULL;
	volatile GrafPtr port = NULL;
	FrameHandle frame = NULL;
	Rect view = *bounds;
	
	require(WinValid(window));
	require(RectWidth(bounds) > 2 * TX_MARGIN);
	require(RectHeight(bounds) > 2 * TX_MARGIN);
	require(TxKindValid(kind));
	TRY {
		GetPort(&port);
		SetPort(window);
		text = HandleBeginClear(sizeof(TextType));
		InsetRect(&view, TX_MARGIN + FRAME_WIDTH, TX_MARGIN + FRAME_WIDTH);
		if (kind == TX_EXTENDED_KIND) {
			TE32KHandle te = TE32KBegin(&view);
			(**text).te = te;
			(**text).kind = kind;
			(**text).maxlength = TE32K_MAX_LENGTH;
		}
		else {
			TEHandle te = TEBegin(&view, kind == TX_STYLED_KIND);
			(**text).te = te;
			(**text).kind = kind;
			(**text).maxlength = TE_MAX_LENGTH;
		}
		frame = FrameBegin(window, bounds, TX_MARGIN);
		(**text).frame = frame;
		(**text).autoview = true;
		(**text).autoredraw = true;
		FrameCursor(frame, iBeamCursor);
		TxChanged(text);
		TxWrapSet(text, false);
		TxHookPasteSet(text, TxHookDefaultPaste);
		TxHookClickLoopSet(text, TxHookDefaultClickLoop);
		WinRegister(window, text, TxEventTable());
		WinTabRegister(window, text);
	} CLEANUP {
		if (port) SetPort(port);
	} CATCH {
		TxEnd(text);
	} ENDTRY;
	ensure(TxValid(text));
	return(text);
}

TextHandle TxBeginEdit(WindowPtr window, const Rect *view)
{
	return(TxBegin(window, view, TX_UNSTYLED_KIND));
}

TextHandle TxBeginStyle(WindowPtr window, const Rect *view)
{
	return(TxBegin(window, view, TX_STYLED_KIND));
}

TextHandle TxBeginExtend(WindowPtr window, const Rect *view)
{
	return(TxBegin(window, view, TX_EXTENDED_KIND));
}

/* limits */

TextIntType TxMaxLines(TextHandle text)
{
	return(TEMMaxLines(text));
}

TextIntType TxMaxLength(TextHandle text)
{
	return((**text).maxlength);
}

void TxMaxLengthSet(TextHandle text, TextIntType maxlength)
{
	require(0 <= maxlength && maxlength <= TEMMaxLength(text));
	/* truncate extra characters */
	if (TxLength(text) > maxlength) {
    	/* find start of character */
		short start = TxSelStart(text);
		short end = TxSelEnd(text);
    	short last = maxlength;
    	char *p = *TxText(text);
		for (last = maxlength; last > 0; last--) {
			short cb = CharByte(p, last); /* CharByte won't move mem */
			if (cb == smSingleByte || cb == smFirstByte)
				break;
		}
		/* truncate at start of character */
		TxSelEndSet(text, TxLength(text));
		TxSelStartSet(text, last);
		TxDelete(text);
		TxSelStartSet(text, min(last, start));
		TxSelEndSet(text, min(last, end));
	}
	(**text).maxlength = maxlength;
	ensure(TxValid(text));
}

/* selection */

TextIntType TxSelStart(TextHandle text)
{
	return(TEMSelStart(text));
}

void TxSelStartSet(TextHandle text, TextIntType start)
{
	require(0 <= start && start <= TxSelEnd(text));
	TEMSelStartSet(text, start);
}

TextIntType TxSelEnd(TextHandle text)
{
	return(TEMSelEnd(text));
}

void TxSelEndSet(TextHandle text, TextIntType end)
{
	require(TxSelStart(text) <= end && end <= TxLength(text));
	TEMSelEndSet(text, end);
}

void TxSelect(TextHandle text, TextIntType start, TextIntType end)
{
	require(TxOffsetValid(text, start));
	require(TxOffsetValid(text, end));
	require(start <= end);
	TEMSelect(text, start, end);
}

void TxSelectLine(TextHandle text, TextIntType line)
{
	TxSelect(text, TxLineStart(text, line), TxLineEnd(text, line));
}

void TxSelectAll(TextHandle text)
{
	TxSelect(text, 0, TxLength(text));
}

/* get width and height of character at specified offset */
static void TxGetWidthAndHeight(TextHandle text, TextIntType offset,
	short *width, short *height)
{
	GrafPtr port = NULL;
	TextState svtext;
	TextStyle style;
	short ascent = 0;
	
	GetPort(&port);
	SetPort(TxPort(text));
	GetTextState(&svtext);
	TxStyleOffset(text, offset, &style, height, &ascent);
	TextFont(style.tsFont);
	TextFace(style.tsFace);
	TextSize(style.tsSize);
	*width = CharWidth(TxCharacter(text, offset)); /* program_note: won't work with multi-byte characters */
	SetTextState(&svtext);
	SetPort(port);
}

/* get rectangle enclosing character at specified offset */
static void TxCharRect(TextHandle text, TextIntType offset, Rect *charRect)
{
	short height;
	short width;
	LPointType pt;

	TxGetWidthAndHeight(text, offset, &width, &height);
	TxPoint(text, offset, &pt);
	charRect->top = pt.v - height;
	charRect->left = pt.h;
	charRect->bottom = pt.v;
	charRect->right = pt.h + width;
}

/* return amount to scroll to show character at the specified offset */
static void TxOffsetShow(TextHandle text, TextIntType offset, LPointType *delta)
{
	Rect view, sel;
	
	TxView(text, &view);
	TxCharRect(text, offset, &sel);
	delta->h = delta->v = 0;
	sel.left--; sel.right++;
	if (sel.bottom > view.bottom) delta->v = view.bottom - sel.bottom;
	if (sel.right > view.right) delta->h = view.right - sel.right;
	if (sel.top < view.top) delta->v = view.top - sel.top;
	if (sel.left < view.left) delta->h = view.left - sel.left;
}

void TxSelectionShow(TextHandle text)
{
	LPointType delta;
	
	TxOffsetShow(text, TxSelStart(text), &delta);
	TxScroll(text, delta.h, delta.v);
}

/* scrolling */

/* call after text has changed to adjust scroll offset and show character
	at specified offset */
void TxAdjust(TextHandle text, TextIntType char_offset)
{
	LPointType all;		/* width and height of all of text */
	LPointType offset;	/* current scroll offset */
	LPointType delta;		/* amount to scroll text */
	LPointType show;		/* amount to scroll text to show char_offset */
	Rect view;				/* text's view rectangle */
	
	require(char_offset == TX_BAD_OFFSET || TxOffsetValid(text, char_offset));
	if (char_offset != TX_BAD_OFFSET)
		TxOffsetShow(text, char_offset, &show);
	else
		show.h = show.v = 0;
	if (show.h >= 0 || show.v >= 0) {
		TxView(text, &view);
		all.h = TxWidthAll(text);
		all.v = TxHeightAll(text);
		offset.h = TxOffsetHorizontal(text);
		offset.v = TxOffsetVertical(text);
		delta.h = min(offset.h, max(RectWidth(&view) + offset.h - all.h, 0));
		delta.v = min(offset.v, max(RectHeight(&view) + offset.v - all.v, 0));
		check(delta.h >= 0 && delta.v >= 0);
		delta.h = (show.h >= 0 ? max(delta.h, show.h) : show.h);
		delta.v = (show.v >= 0 ? max(delta.v, show.v) : show.v);
	}
	else
		delta = show;
	TxScroll(text, delta.h, delta.v);
}

/* called after text changes */
void TxChanged(TextHandle text)
{
	require(TxValid(text));
	(**text).size.h = (**text).size.v = TX_BAD_SIZE;
	if ((**text).autoview)
		TxAdjust(text, TxSelStart(text));
}

void TxScroll(TextHandle text, TextIntType dh, TextIntType dv)
{
	LRectType dest;
	Rect view;
	
	TxView(text, &view);
	TxDestination(text, &dest);
	check(dest.top <= view.top);
	check(dest.left <= view.left);
	dv = min(dv, view.top - dest.top);
	dh = min(dh, view.left - dest.left);
	dv = max(dv, min(0, view.top - dest.top + RectHeight(&view) - TxHeightAll(text)));
	dh = max(dh, min(0, view.left - dest.left + RectWidth(&view) - TxWidthAll(text)));
	if (dh || dv) /* test fixes bug in system 4.2 */
		TEMScroll(text, dh, dv);
}

/* text buffer */

Handle TxText(TextHandle text)
{
	return(TEMText(text));
}

void TxReplaceText(TextHandle text, Handle h)
{
	require(HandleSize(h) <= TxMaxLength(text));
	TEMReplaceText(text, h);
	TxChanged(text);
}

void TxSetText(TextHandle text, const char *p, TextIntType length)
{
	require(0 <= length && length <= TxMaxLength(text));
	TEMSetText(text, p, length);
	TxChanged(text);
}

TextIntType TxGetText(TextHandle text, char *p, TextIntType length)
{
	require(length >= 0);
	length = min(length, TxLength(text));
	BlockMove(*TxText(text), p, length);
	return(length);
}

/* characters */

char TxCharacter(TextHandle text, TextIntType offset)
{
	require(TxOffsetValid(text, offset));
	return((*TxText(text))[offset]);
}

TextIntType TxLength(TextHandle text)
{
	return(TEMLength(text));
}

void TxPoint(TextHandle text, TextIntType offset, LPointType *lpt)
{
	require(TxOffsetValid(text, offset));
	TEMGetPoint(text, offset, lpt);
}

TextIntType TxOffset(TextHandle text, const LPointType *lpt)
{
	return(TEMGetOffset(text, lpt));
}

/* lines */

TextIntType TxLines(TextHandle text)
{
	return(TEMLines(text));
}

TextIntType TxLineStart(TextHandle text, TextIntType line)
{
	require(TxLineValid(text, line));
	return(TEMLineStart(text, line));
}

TextIntType TxLineEnd(TextHandle text, TextIntType line)
{
	require(TxLineValid(text, line));
	return(TEMLineEnd(text, line));
}

/* return the line where the selection starts */
TextIntType TxLineWithOffset(TextHandle text, TextIntType offset)
{
	TextIntType i = 0;
	TextIntType nlines = 0;
	TextIntType start = 0;
	
	require(TxOffsetValid(text, offset));
	nlines = TxLines(text);
	for (i = 0; i < nlines && TxLineStart(text, i) < offset; i++)
		;
	if (i > 0)
		--i;
	ensure(TxLineValid(text, i));
	return(i);
}

/* return the length of the specified line */
TextIntType TxLineLength(TextHandle text, TextIntType line)
{
	return(TxLineEnd(text, line) - TxLineStart(text, line));
}

/* get the text of the specified line */
void TxLineGet(TextHandle text, TextIntType line, char *s, TextIntType len)
{
	TextIntType start = 0;
	
	require(len > 0);
	start = TxLineStart(text, line);
	len = min(len - 1, TxLineLength(text, line));
	BlockMove(*TxText(text) + start, s, len);
	s[len] = 0;
}

/* rectangles */

void TxView(TextHandle text, Rect *view)
{
	TEMView(text, view);
	ensure(RectValid(view));
}

void TxViewSet(TextHandle text, const Rect *view)
{
	require(RectValid(view));
	TEMViewSet(text, view);
}

void TxDestination(TextHandle text, LRectType *ldest)
{
	TEMDest(text, ldest);
	ensure(LRectValid(ldest));
}

void TxDestinationSet(TextHandle text, const LRectType *ldest)
{
	require(LRectValid(ldest));
	TEMDestSet(text, ldest);
	TxChanged(text);
}

/* return horizontal offset of text (i.e., amount text has been scrolled) */
TextIntType TxOffsetHorizontal(TextHandle text)
{
	Rect view;
	LRectType dest;
	
	TxView(text, &view);
	TxDestination(text, &dest);
	check(dest.left <= view.left);
	return(view.left - dest.left);
}

/* return vertical offset of text (i.e., amount text has been scrolled) */
TextIntType TxOffsetVertical(TextHandle text)
{
	Rect view;
	LRectType dest;
	
	TxView(text, &view);
	TxDestination(text, &dest);
	check(dest.top <= view.top);
	return(view.top - dest.top);
}

/* justification and word wrap */

short TxJust(TextHandle text)
{
	return(TEMJust(text));
}

Boolean TxWrap(TextHandle text)
{
	return(TEMWrap(text));
}

/* return word wrap margin */
TextIntType TxMargin(TextHandle text)
{
	LRectType dest;

	TxDestination(text, &dest);
	return(TxWrap(text) ? LRectWidth(&dest) : TX_MAX_MARGIN);
}

void TxJustSet(TextHandle text, short just)
{
	TEMJustSet(text, just);
	TxChanged(text);
}

void TxWrapSet(TextHandle text, Boolean wrap)
{
	TEMWrapSet(text, wrap);
	TxChanged(text);
}

/* set word wrap margin; if 0 then wrap to view rectangle */
void TxMarginSet(TextHandle text, TextIntType margin)
{
	Rect view;
	LRectType dest;
	
	require(0 <= margin && margin <= TX_MAX_MARGIN);
	TxView(text, &view);
	TxDestination(text, &dest);
	if (margin) {
		if (dest.right > TX_MAX_MARGIN - margin) /* checks for overflow */
			dest.right = dest.left + TX_MAX_MARGIN;
		else
			dest.right = dest.left + margin;
	}
	else
		dest.right = dest.left + RectWidth(&view);
	TxDestinationSet(text, &dest);
	TxChanged(text);
	ensure(TxValid(text));
}

/* style */

short TxFont(TextHandle text)
{
	TextStyle style;
	short height, ascent;
	
	TxStyle(text, &style, &height, &ascent);
	return(style.tsFont);
}

short TxSize(TextHandle text)
{
	TextStyle style;
	short height, ascent;
	
	TxStyle(text, &style, &height, &ascent);
	return(style.tsSize);
}

Style TxFace(TextHandle text)
{
	TextStyle style;
	short height, ascent;
	
	TxStyle(text, &style, &height, &ascent);
	return(style.tsFace);
}

static void FixStyle(short mode, TextStyle *style)
{
	if ((mode & doSize) != 0 && style->tsSize == 0)
		style->tsSize = GetDefFontSize();
}

void TxStyleContinuous(TextHandle text, short *mode, TextStyle *style)
{
	TEMStyleContinuous(text, mode, style);
	FixStyle(*mode, style);
}

void TxStyleOffset(TextHandle text, TextIntType offset, TextStyle *style,
	short *height, short *ascent)
{
	require(TxOffsetValid(text, offset));
	TEMStyleOffset(text, offset, style, height, ascent);
	FixStyle(doAll, style);
}

void TxStyle(TextHandle text, TextStyle *style, short *height, short *ascent)
{
	TEMStyle(text, style, height, ascent);
	FixStyle(doAll, style);
}

void TxStyleSet(TextHandle text, short mode, const TextStyle *style)
{
	TEMStyleSet(text, mode, style, (**text).autoredraw);
	TxChanged(text);
}

StScrpHandle TxStyleScrap(TextHandle text)
{
	return(TEMStyleScrap(text));
}

void TxStyleScrapSet(TextHandle text, StScrpHandle styles)
{
	require(! styles || HandleValidSize(styles, sizeof(short) + sizeof(ScrpSTElement)));
	TEMStyleScrapSet(text, styles, (**text).autoredraw);
	TxChanged(text);
}

/* dimensions */

TextIntType TxWidth(TextHandle text, TextIntType start, TextIntType end)
{
	TextIntType width;
	
	require(TxLineValid(text, start));
	require(start <= end && end <= TxLines(text) + 1);
	if (TxWrap(text))
		width = TxMargin(text);
	else {
		short final;
		const TicksType timeout = TICKS_SEC / 4;
		width = TEMWidth(text, start, end, timeout, &final);
		if (final < end && width < TX_MAX_MARGIN) /* timed out */
			width = TX_MAX_MARGIN;
	}
	return(width);
}

TextIntType TxWidthAll(TextHandle text)
{
	require(TxValid(text));
	if ((**text).size.h == TX_BAD_SIZE) {
		TextIntType width = TxWidth(text, 0, TxLines(text));
		(**text).size.h = width;
	}
	return((**text).size.h);
}

TextIntType TxHeight(TextHandle text, TextIntType start, TextIntType end)
{
	require(TxLineValid(text, start));
	require(start <= end && end <= TxLines(text) + 1);
	return(TEMHeight(text, start, end));
}

TextIntType TxHeightAll(TextHandle text)
{
	require(TxValid(text));
	if ((**text).size.v == TX_BAD_SIZE) {
		TextIntType height = TxHeight(text, 0, TxLines(text));
		(**text).size.v = height;
	}
	return((**text).size.v);
}

/* edit */

void TxCut(TextHandle text)
{
	TxCopy(text);
	TxDelete(text);
}

void TxCopy(TextHandle text)
{
	TEMCopy(text);
}

void TxPaste(TextHandle text)
{
	if (CanPaste(text)) {
		TEMPaste(text);
		TxChanged(text);
	}
	else
		SysBeep(5);
}

void TxDelete(TextHandle text)
{
	TEMDelete(text);
	TxChanged(text);
}

void TxKey(TextHandle text, char key)
{
	Boolean enter = true;
	Boolean changed = true;
	
	if (KeyIsArrow(key))
		changed = false;
	else if (key == backspaceKey && TxSelStart(text) == TxSelEnd(text) &&
				TxSelStart(text) == 0)
	{
		changed = false;
	}
	if (! KeyIsArrow(key) && ((**text).flags & TX_READ_ONLY) != 0)
		enter = false;
	else if (key != backspaceKey && ! KeyIsArrow(key)) {
		if (TxSelStart(text) == TxSelEnd(text) &&
			 TxLength(text) >= TxMaxLength(text))
		{
			enter = false;
		}
		else if ((**text).hooks.key &&
			! (**text).hooks.key(text, key, (**text).hooks.data))
		{
			enter = false;
		}
	}
	if (enter) {
		if (changed && key != backspaceKey)
			check(TxLength(text) - TxSelEnd(text) + TxSelStart(text) + 1 <= TxMaxLength(text));
		TEMKey(text, key);
		if (changed) {
			if ((**text).autoview && key != backspaceKey &&
				 ! KeyIsArrow(key) && TxSelStart(text) > 0)
			{
				/* show the character that was just entered */
				(**text).autoview = false;
				TxChanged(text);
				(**text).autoview = true;
				TxAdjust(text, TxSelStart(text) - 1);
			}
			else
				TxChanged(text);
		}
		else if ((**text).autoview)
			TxSelectionShow(text);
	}
	else
		SysBeep(5);
}

void TxInsert(TextHandle text, const char *p, TextIntType length,
	StScrpHandle style)
{
	require(! style || HandleValidSize(style, sizeof(short) + sizeof(ScrpSTElement)));
	require(p != NULL);
	require(0 <= length && length + TxLength(text) <= TxMaxLength(text));
	TEMInsert(text, p, length, style);
	TxChanged(text);
}

/* scrap */

void TxScrapToDeskScrap(void)
{
}

void TxScrapFromDeskScrap(void)
{
}

TextIntType TxScrapLength(void)
{
	return(ScrapLength('TEXT'));
}

/* idle */

void TxIdle(TextHandle text)
{
	TEMIdle(text);
}

/* activate */

Boolean TxActive(TextHandle text)
{
	return(TEMActive(text));
}

void TxActivate(TextHandle text, Boolean active)
{
	require(TxValid(text));
	FrameActivate((**text).frame, active);
	TEMActivate(text, FrameHilite((**text).frame));
}

void TxFocus(TextHandle text, Boolean focus)
{
	require(TxValid(text));
	FrameFocus((**text).frame, focus);
	TEMActivate(text, FrameHilite((**text).frame));
}

/* draw */

GrafPtr TxPort(TextHandle text)
{
	return(TEMPort(text));
}

void TxPortSet(TextHandle text, GrafPtr port)
{
	TEMPortSet(text, port);
}

void TxInval(TextHandle text)
{
	Rect view;
	GrafPtr port;
	
	TxView(text, &view);
	GetPort(&port);
	SetPort(TxPort(text));	
	InvalRect(&view);
	SetPort(port);
}

void TxUpdate(TextHandle text)
{
	Rect view;
	GrafPtr port = NULL;

	GetPort(&port);
	SetPort(TxPort(text));
	TxView(text, &view);
	EraseRect(&view);
	TEMUpdate(text, &view);
	SetPort(port);
}

/* mouse down */

void TxClick(TextHandle text, Point pt, Boolean extend)
{
	require(! gClickText);
	if (! gClickRgn) {
		gClickRgn = NewRgn();
		FailNIL(gClickRgn);
	}
	if (((**text).flags & TX_IGNORE_CLICK) == 0) {
		gClickText = text;
		TEMClick(text, pt, extend);
	}
	gClickText = NULL;
	SetEmptyRgn(gClickRgn);
}

Boolean TxWithin(TextHandle text, Point where)
{
	Rect view;
	
	TxView(text, &view);
	return(PtInRect(where, &view));
}

void TxMouseDown(TextHandle text, EventRecord *event)
{
	Point where = event->where;

	GlobalToPort(&where, TxPort(text));
	TxClick(text, where, (event->modifiers & shiftKey) != 0);
}

/* adjust */

void TxAdjustMenu(TextHandle text)
{	
	Boolean act = TxActive(text);
	Boolean sel = (TxSelStart(text) < TxSelEnd(text));
	Boolean rdwr = (((**text).flags & TX_READ_ONLY) == 0);
	
	MenuCmdEnable(CMD_EDIT);
	if (sel && act && rdwr) MenuCmdEnable(CMD_CUT);
	if (sel && act && rdwr) MenuCmdEnable(CMD_CLEAR);
	if (sel && act) MenuCmdEnable(CMD_COPY);
	if (TxScrapLength() > 0 && rdwr && CanPaste(text)) MenuCmdEnable(CMD_PASTE);
	if (act) MenuCmdEnable(CMD_SELECT_ALL);
}

TicksType TxAdjustSleep(TextHandle text)
{
	TicksType sleep = LONG_MAX;
	
	require(! text || TxValid(text));
	if (text && TxActive(text) && TxSelStart(text) == TxSelEnd(text))
		sleep = GetCaretTime();
	return(sleep);
}

/* menu */

Boolean TxMenu(TextHandle text, const MenuPickType *pick)
{
	Boolean handled = true;
	
	switch (pick->cmd) {
	case CMD_CUT:			TxCut(text); break;
	case CMD_COPY:			TxCopy(text); break;
	case CMD_PASTE:		TxPaste(text); break;
	case CMD_CLEAR:		TxDelete(text); break;
	case CMD_SELECT_ALL:	TxSelectAll(text); break;
	default:
		handled = false;
		break;
	}
	return(handled);
}

/* position and size */

/* resize the text */
void TxResize(TextHandle text, short dh, short dv)
{
	Rect view;
	LRectType dest;
	
	require(TxValid(text));
	if ((FrameFlags((**text).frame) & FRAME_ANCHOR_RIGHT) == 0) dh = 0;
	if ((FrameFlags((**text).frame) & FRAME_ANCHOR_BOTTOM) == 0) dv = 0;
	TxView(text, &view);
	view.right += dh;
	view.bottom += dv;
	TxViewSet(text, &view);
	if (((**text).flags & TX_WRAP_TO_VIEW) != 0) {
		TxDestination(text, &dest);
		dest.right += dh;
		dest.bottom += dv;
		TxDestinationSet(text, &dest);
		TxInval(text);
		TxCalculate(text);
	}
	TxAdjust(text, TX_BAD_OFFSET);
}

/* flags */

TextFlagsType TxFlags(TextHandle text)
{
	require(TxValid(text));
	return((**text).flags);
}

void TxFlagsSet(TextHandle text, TextFlagsType flags)
{
	require(TxValid(text));
	(**text).flags = flags;
	if ((**text).flags & TX_READ_ONLY)
		FrameCursor((**text).frame, 0);
	else
		FrameCursor((**text).frame, iBeamCursor);
}

/* hooks */

static pascal Boolean clickloop(void)
{
	Point where;
	Rect portRect;
	GrafPtr port = NULL;
	TextHandle text = gClickText;
	RgnHandle clip = gClickRgn;
	Boolean result = false;
	
	asm { move.l d2, -(sp) } /* see IM-1, p395 */
	require(TxValid(text));
	if ((**text).hooks.clickloop) {
		GetPort(&port);
		SetPort(TxPort(text));
		WinPortRect(TxPort(text), &portRect);
		GetClip(clip);
		ClipRect(&portRect);
		GetMouse(&where);
		result = (**text).hooks.clickloop(text, where, (**text).hooks.data);
		SetClip(clip);
		SetPort(port);
	}
	asm { move.l (sp)+, d2 }
	return(result);
	
}

void TxHooks(TextHandle text, TextHooksType *hooks)
{
	require(TxValid(text));
	*hooks = (**text).hooks;
}

void TxHooksSet(TextHandle text, const TextHooksType *hooks)
{
	require(TxValid(text));
	(**text).hooks = *hooks;
}

void TxHookDataSet(TextHandle text, void *data)
{
	require(TxValid(text));
	(**text).hooks.data = data;
}

void TxHookKeySet(TextHandle text, TextHookKeyType hook)
{
	require(TxValid(text));
	(**text).hooks.key = hook;
}

void TxHookPasteSet(TextHandle text, TextHookPasteType hook)
{
	require(TxValid(text));
	(**text).hooks.paste = hook;
}

void TxHookClickLoopSet(TextHandle text, TextHookClickLoopType hook)
{
	require(TxValid(text));
	(**text).hooks.clickloop = hook;
	TEMClickLoopSet(text, clickloop);
}

/* scroll the text when the mouse is dragged outside of the text */
Boolean TxHookDefaultClickLoop(TextHandle text, Point mouse, void *data)
{
	Rect		view;					/* the view rectangle */
	short		dh = 0, dv = 0;	/* amount to scroll */
	Boolean	slow = false;		/* true if should scroll slowly */
		
	/* determine the amount to scroll in the direction of the mouse
		(in relation to the view rectangle) */
	TxView(text, &view);
	if (mouse.v < view.top) {
		dv = TX_DELTAV;
		if (! slow) slow = (view.top - mouse.v < TX_SLOW_MARGIN);
	}
	else if (mouse.v > view.bottom) {
		dv = -TX_DELTAV;
		if (! slow) slow = (mouse.v - view.bottom < TX_SLOW_MARGIN);
	}
	if (mouse.h < view.left) {
		dh = TX_DELTAH;
		if (! slow) slow = (view.left - mouse.h < TX_SLOW_MARGIN);
	}
	else if (mouse.h > view.right) {
		dh = -TX_DELTAH;
		if (! slow) slow = (mouse.h - view.right < TX_SLOW_MARGIN);
	}
	if (dh || dv) {
		/* Delay for several ticks to achieve a slow scroll speed if the mouse
			is dragged only a few pixels outside of the view rectangle. This
			is helpful for people with relatively fast machines. */
		if (slow) {
			long time;
			Delay(TX_SLOW_DELAY, &time);
		}
		TxScroll(text, dh, dv);
	}
	return(true);
}

/* default paste hook verifies that scrap isn't too large and
	calls key hook for each character in scrap */
Boolean TxHookDefaultPaste(TextHandle text, void *data)
{
	register long offset, length;
	register Handle scrap = NULL;
	register Boolean result = true;
	SignedByte state = 0;
	
	if ((**text).hooks.key) {
		length = ScrapLength('TEXT');
		if (length > 0) {
			scrap = ScrapHandle();
			if (scrap) {
				offset = ScrapOffset('TEXT');
				length += offset;
				check(length <= HandleSize(scrap));
				state = HandleNoPurge(scrap);
				while (result && offset < length)
					result = (**text).hooks.key(text, (*scrap)[offset++], (**text).hooks.data);
				HandleRestore(scrap, state);
			}
		}
	}
	return(result);
}

#include <ctype.h> /* program_note: should use international utilities */

/* key hook for a field containing an unsigned integer, only allows
	digits to be entered */
Boolean TxHookKeyUnsignedInteger(TextHandle text, char key, void *data)
{
	return(CharByte(&key, 0) == smSingleByte && isdigit(key));
}

/* key hook for a field containing an unsigned floating point value,
	only allows digits and a '.' to be entered */
Boolean TxHookKeyUnsignedFloat(TextHandle text, char key, void *data)
{
	static char decimalPt;
	Boolean result = false;
	
	if (! decimalPt) {
		/* get decimal point character from international resource */
		decimalPt = (**(Intl0Hndl) IUGetIntl(0)).decimalPt;
		check(CharByte(&decimalPt, 0) == smSingleByte);
	}
	if (key == decimalPt) {
		/* disallow if text already contains a decimal point */
		TextIntType start = TxSelStart(text);
		TextIntType end = TxSelEnd(text);
		TextIntType length = TxLength(text);
		register char *p = StripAddress(*TxText(text));
		register char *last = p + length;
		char *selstart = p + start;
		char *selend = p + end;
		
		while (p < last) {
			if (*p == decimalPt) {
				if (start == end || p < selstart || selend < p)
					break;
			}
			p++;
		}
		result = (p == last);
	}
	else
		result = TxHookKeyUnsignedInteger(text, key, data);
	return(result);
}

/* miscelaneous */

void TxCalculate(TextHandle text)
{
	TEMCalText(text);
	TxChanged(text);
}

void TxAutoViewSet(TextHandle text, Boolean autoview)
{
	require(TxValid(text));
	(**text).autoview = autoview;
}

Boolean TxAutoRedraw(TextHandle text)
{
	require(TxValid(text));
	return((**text).autoredraw);
}

void TxAutoRedrawSet(TextHandle text, Boolean autoredraw)
{
	require(TxValid(text));
	(**text).autoredraw = autoredraw;
}
