#pragma once

#include <limits.h>
#include "FrameLib.h"
#include "MenuLib.h"
#include "TimeLib.h"
#include "LongPointLib.h"
#include "LongRectangleLib.h"

/* constants */
#define TX_MARGIN			(2)			/* margin between text and frame */
#define TX_MIN_FONT		(1)			/* minimum font size */
#define TX_MAX_FONT		SHRT_MAX		/* maximum font size */
#define TX_MIN_MARGIN	(20)			/* minimum wrap margin */
#define TX_MAX_MARGIN	(5184)		/* maximum wrap margin (72 inches) */
#define TX_BAD_OFFSET	(-1L)			/* illegal character offset */
#define TX_BAD_SIZE		(-1L)			/* illegal text size */
#define TX_DELTAV			(16)			/* number of pixels to scroll vertically */
#define TX_DELTAH			(8)			/* number of pixels to scroll horizontally */

/* flags that modify the behavior of the text */
#define TX_READ_ONLY		(1<<0)		/* can't modify text */
#define TX_IGNORE_CLICK	(1<<1)		/* ignore clicks */															 
#define TX_WRAP_TO_VIEW	(1<<2)		/* wrap to width of view rect */

/* type definitions */

typedef unsigned short TextFlagsType;
typedef long TextIntType;

typedef enum {
	TX_UNSTYLED_KIND,		/* original unstyled TextEdit record */
	TX_STYLED_KIND,		/* new styled TextEdit record */
	TX_EXTENDED_KIND,		/* extended record (i.e., TE32K) */
	TX_LAST_KIND
} TextKindType;

typedef struct TextStruct TextType, *TextPtr, **TextHandle;

/* hook routines */
typedef Boolean (*TextHookClickLoopType)(TextHandle text, Point where, void *data);
typedef Boolean (*TextHookKeyType)(TextHandle text, char key, void *data);
typedef Boolean (*TextHookPasteType)(TextHandle text, void *data);

/* hooks for text records */
typedef struct {
	TextHookClickLoopType clickloop;	/* click loop routine */
	TextHookKeyType key;			/* key hook, returns true if can enter char */
	TextHookPasteType paste;	/* paste hook, returns true if can paste scrap */
	void *data;						/* data for hooks */
} TextHooksType;

struct TextStruct {
	TextKindType	kind;			/* type of text edit record */
	TextFlagsType	flags;		/* various flags */
	TextHooksType	hooks;		/* hook routines */
	TextIntType		maxlength;	/* maximum length of text */
	FrameHandle		frame;		/* frame surrounding text */
	LPointType		size;			/* size of text */
	Boolean			autoview;	/* if true, selection is automatically shown */
	Boolean			autoredraw;	/* if true, style changes recalculate and redraw text */
	void				*te;			/* the text edit record */
};

/* validation */
Boolean TxValid(TextHandle text);
Boolean TxLineValid(TextHandle text, TextIntType line);
Boolean TxOffsetValid(TextHandle text, TextIntType offset);

/* allocation */
void TxEnd(TextHandle text);
TextHandle TxBegin(WindowPtr window, const Rect *bounds, TextKindType kind);
TextHandle TxBeginEdit(WindowPtr window, const Rect *bounds);
TextHandle TxBeginStyle(WindowPtr window, const Rect *bounds);
TextHandle TxBeginExtend(WindowPtr window, const Rect *bounds);

/* limits */
TextIntType TxMaxLines(TextHandle text);
TextIntType TxMaxLength(TextHandle text);
void TxMaxLengthSet(TextHandle text, TextIntType maxlength);

/* selection */
TextIntType TxSelEnd(TextHandle text);
TextIntType TxSelStart(TextHandle text);
void TxSelEndSet(TextHandle text, TextIntType end);
void TxSelStartSet(TextHandle text, TextIntType start);
void TxSelect(TextHandle text, TextIntType start, TextIntType end);
void TxSelectLine(TextHandle text, TextIntType line);
void TxSelectAll(TextHandle text);
void TxSelectionShow(TextHandle text);

/* scrolling */
void TxAdjust(TextHandle text, TextIntType char_offset);
void TxChanged(TextHandle text);
void TxScroll(TextHandle text, TextIntType dh, TextIntType dv);

/* text buffer */
Handle TxText(TextHandle text);
void TxReplaceText(TextHandle text, Handle h);
void TxReplaceTextAndStyle(TextHandle text, Handle h, StScrpHandle scrp);
void TxSetText(TextHandle text, const char *p, TextIntType length);
TextIntType TxGetText(TextHandle text, char *p, TextIntType length);

/* characters */
TextIntType TxLength(TextHandle text);
char TxCharacter(TextHandle text, TextIntType offset);
void TxPoint(TextHandle text, TextIntType offset, LPointType *pt);
TextIntType TxOffset(TextHandle text, const LPointType *pt);

/* lines */
TextIntType TxLines(TextHandle text);
TextIntType TxLineStart(TextHandle text, TextIntType line);
TextIntType TxLineEnd(TextHandle text, TextIntType line);
TextIntType TxLineWithOffset(TextHandle text, TextIntType offset);
TextIntType TxLineLength(TextHandle text, TextIntType line);
void TxLineGet(TextHandle text, TextIntType line, char *s, TextIntType max);

/* rectangles */
void TxView(TextHandle text, Rect *view);
void TxViewSet(TextHandle text, const Rect *view);
void TxDestination(TextHandle text, LRectType *destination);
void TxDestinationSet(TextHandle text, const LRectType *destination);
TextIntType TxOffsetHorizontal(TextHandle text);
TextIntType TxOffsetVertical(TextHandle text);

/* justification and word wrap */
short TxJust(TextHandle text);
Boolean TxWrap(TextHandle text);
TextIntType TxMargin(TextHandle text);
void TxJustSet(TextHandle text, short just);
void TxWrapSet(TextHandle text, Boolean wrap);
void TxMarginSet(TextHandle text, TextIntType margin);

/* style */
short TxFont(TextHandle text);
short TxSize(TextHandle text);
Style TxFace(TextHandle text);
void TxStyleContinuous(TextHandle text, short *modes, TextStyle *style);
void TxStyleOffset(TextHandle text, TextIntType offset, TextStyle *style,
	short *height, short *ascent);
void TxStyle(TextHandle text, TextStyle *style, short *height, short *ascent);
void TxStyleSet(TextHandle text, short modes, const TextStyle *style);
StScrpHandle TxStyleScrap(TextHandle text);
void TxStyleScrapSet(TextHandle text, StScrpHandle styles);

/* dimensions */
TextIntType TxWidth(TextHandle text, TextIntType start, TextIntType end);
TextIntType TxWidthAll(TextHandle text);
TextIntType TxHeight(TextHandle text, TextIntType start, TextIntType end);
TextIntType TxHeightAll(TextHandle text);

/* edit */
void TxCut(TextHandle text);
void TxCopy(TextHandle text);
void TxPaste(TextHandle text);
void TxDelete(TextHandle text);
void TxKey(TextHandle text, char key);
void TxInsert(TextHandle text, const char *p, TextIntType length,
	StScrpHandle style);

/* scrap */
void TxScrapToDeskScrap(void);
void TxScrapFromDeskScrap(void);
TextIntType TxScrapLength(void);

/* idle */
void TxIdle(TextHandle text);

/* activate */
Boolean TxActive(TextHandle text);
void TxActivate(TextHandle text, Boolean active);
void TxFocus(TextHandle text, Boolean focus);

/* draw */
GrafPtr TxPort(TextHandle text);
void TxPortSet(TextHandle text, GrafPtr port);
void TxInval(TextHandle text);
void TxUpdate(TextHandle text);

/* mouse down */
void TxClick(TextHandle text, Point pt, Boolean extend);
Boolean TxWithin(TextHandle text, Point where);
void TxMouseDown(TextHandle text, EventRecord *event);

/* adjust */
void TxAdjustMenu(TextHandle text);
TicksType TxAdjustSleep(TextHandle text);
Boolean TxAdjustCursor(TextHandle text, Point where, RgnHandle cursorRgn);

/* menu */
Boolean TxMenu(TextHandle text, const MenuPickType *pick);

/* position and size */
void TxResize(TextHandle text, short dh, short dv);

/* flags */
TextFlagsType TxFlags(TextHandle text);
void TxFlagsSet(TextHandle text, TextFlagsType flags);

/* hooks */
void TxHooks(TextHandle text, TextHooksType *hooks);
void TxHooksSet(TextHandle text, const TextHooksType *hooks);
void TxHookDataSet(TextHandle text, void *data);
void TxHookKeySet(TextHandle text, TextHookKeyType hook);
void TxHookPasteSet(TextHandle text, TextHookPasteType hook);
void TxHookClickLoopSet(TextHandle text, TextHookClickLoopType hook);
Boolean TxHookDefaultClickLoop(TextHandle text, Point where, void *data);
Boolean TxHookDefaultPaste(TextHandle text, void *data);
Boolean TxHookKeyUnsignedInteger(TextHandle text, char key, void *data);
Boolean TxHookKeyUnsignedFloat(TextHandle text, char key, void *data);
Boolean TxHookPasteUnsignedInteger(TextHandle text, void *data);

/* miscelaneous */
void TxCalculate(TextHandle text);
Boolean TxClickLoop(TextHandle text, Point where, void *data);
void TxAutoViewSet(TextHandle text, Boolean autoview);
Boolean TxAutoRedraw(TextHandle text);
void TxAutoRedrawSet(TextHandle text, Boolean autoredraw);

/* event table */
const /* EventTableType */ void *TxEventTable(void);
void TxEventTableRegister(void);
